/*      q_factal.c

task:   read in 1 int, 4 doubles from STDIN/ARGV
        square quaternion, subtrack one
        print to STDOUT
        repeat n times

author: Douglas B Sweetser

copyright 2002 Douglas B Sweetser, sweetser@theworld.com

compiled with: gcc -Wall -o q_fractal q_fractal.c q_lib.c -lm

see GPL license info at the end of this file
*/


/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "q_lib.h"



int main (int argc, char *argv[])
{
  /* variables */
  int i, n_steps;
  double q1[4], q_temp[4], q_out[4];
  double norm, cutoff;


  /* initialize */
  q_one (q1);
  cutoff = 10000;

  /* data input */
  if (argc == 6)
    {
      n_steps = atoi (argv[1]);
      q_out[0] = atof (argv[2]);
      q_out[1] = atof (argv[3]);
      q_out[2] = atof (argv[4]);
      q_out[3] = atof (argv[5]);
    }
    else
    if (argc == 5)
	{
        q_out[0] = atof (argv[1]);
        q_out[1] = atof (argv[2]);
        q_out[2] = atof (argv[3]);
        q_out[3] = atof (argv[4]);
        fscanf (stdin, "%i", &n_steps);
	}
    else
    if (argc == 2)
        {
	  n_steps = atoi (argv[1]);
          q_scan (q_out);
	}
     else
       {
        fscanf (stdin, "%i", &n_steps);
        q_scan (q_out);
       }


  /* action */
  for (i = 0; i < n_steps; i++)
    {
      q_print (q_out);
 
      q_x (q_out, q_out, q_temp);

      q_sub (q_temp, q1, q_out);

      if ((q_out[0] > cutoff) || (q_out[1] > cutoff) || (q_out[2] > cutoff) || (q_out[3] > cutoff) || (-1 * q_out[0] > cutoff) || (-1 * q_out[1] > cutoff) || (-1 * q_out[2] > cutoff) || (-1 * q_out[3] > cutoff) ) {
        i++;
        fprintf (stderr, "%i\n", i);
	break;
      }
      d_norm (q_out, &norm);
      if (norm == 0)
	{	
        fprintf (stderr, "-%i\n", i);
	break;
      }
    }

  /* signals */
  return 0;

}


/*  License

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
