/* q_lib.c

   task: math functions for quaternions that are arrays of 4 doubles
   The results are placed in the last array or double

   author: Douglas B Sweetser <sweetser@world.std.com>

   copywrite: 2002 Douglas B Sweetser

   licence: see GPL license info at the end of this file
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "q_lib.h"

/*  Creation */
void
q_zero (double q1[])
{
  q1[0] = 0.0;
  q1[1] = 0.0;
  q1[2] = 0.0;
  q1[3] = 0.0;
}

void
q_one (double q1[])
{
  q1[0] = 1.0;
  q1[1] = 0.0;
  q1[2] = 0.0;
  q1[3] = 0.0;
}

void
q_construct (double q1[], double t, double x, double y, double z)
{
  q1[0] = t;
  q1[1] = x;
  q1[2] = y;
  q1[3] = z;
}

void
get_q1 (int argc, char *argv[], double q1[])
{
  if (argc == 5)
    {
      q1[0] = atof (argv[1]);
      q1[1] = atof (argv[2]);
      q1[2] = atof (argv[3]);
      q1[3] = atof (argv[4]);
    }
  else
    {
      q_scan (q1);
    }
}

void
get_q1_d (int argc, char *argv[], double q1[], double *d)
{
  if (argc == 6)
    {
      q1[0] = atof (argv[1]);
      q1[1] = atof (argv[2]);
      q1[2] = atof (argv[3]);
      q1[3] = atof (argv[4]);
      *d = atof (argv[5]);
    }
  else 
  if (argc == 5)
    {
      q1[0] = atof (argv[1]);
      q1[1] = atof (argv[2]);
      q1[2] = atof (argv[3]);
      q1[3] = atof (argv[4]);
      fscanf (stdin, "%lf", d);
    }
  else
  if (argc == 2)
    {
      *d = atof (argv[1]);
      q_scan (q1);
    }
  else
    {
      q_scan (q1);
      fscanf (stdin, "%lf", d);
    }
}

void
get_q1_i (int argc, char *argv[], double q1[], int *i)
{
  if (argc == 6)
    {
      q1[0] = atof (argv[1]);
      q1[1] = atof (argv[2]);
      q1[2] = atof (argv[3]);
      q1[3] = atof (argv[4]);
      *i = atoi (argv[5]);
    }
  else 
  if (argc == 5)
    {
      q1[0] = atof (argv[1]);
      q1[1] = atof (argv[2]);
      q1[2] = atof (argv[3]);
      q1[3] = atof (argv[4]);
      fscanf (stdin, "%i", i);
    }
  else
  if (argc == 2)
    {
      *i = atoi (argv[1]);
      q_scan (q1);
    }
  else
    {
      q_scan (q1);
      fscanf (stdin, "%i", i);
    }
}

void
get_q1_q2 (int argc, char *argv[], double q1[], double q2[])
{
  if (argc == 9)
    {
      q1[0] = atof (argv[1]);
      q1[1] = atof (argv[2]);
      q1[2] = atof (argv[3]);
      q1[3] = atof (argv[4]);
      q2[0] = atof (argv[5]);
      q2[1] = atof (argv[6]);
      q2[2] = atof (argv[7]);
      q2[3] = atof (argv[8]);
    }
  else 
  if (argc == 5)
    {
      q2[0] = atof (argv[1]);
      q2[1] = atof (argv[2]);
      q2[2] = atof (argv[3]);
      q2[3] = atof (argv[4]);
      q_scan (q1);
    }
  else
    {
      q_scan (q1);
      q_scan (q2);
    }
}

void
get_q1_q2_i (int argc, char *argv[], double q1[], double q2[], int *i)
{
  if (argc == 10)
    {
      q1[0] = atof (argv[1]);
      q1[1] = atof (argv[2]);
      q1[2] = atof (argv[3]);
      q1[3] = atof (argv[4]);
      q2[0] = atof (argv[5]);
      q2[1] = atof (argv[6]);
      q2[2] = atof (argv[7]);
      q2[3] = atof (argv[8]);
      *i    = atoi (argv[9]);
    }
  else 
  if (argc == 6)
    {
      q2[0] = atof (argv[1]);
      q2[1] = atof (argv[2]);
      q2[2] = atof (argv[3]);
      q2[3] = atof (argv[4]);
      *i    = atoi (argv[5]);
      q_scan (q1);
    }
  else
  if (argc == 5)
    {
      q2[0] = atof (argv[1]);
      q2[1] = atof (argv[2]);
      q2[2] = atof (argv[3]);
      q2[3] = atof (argv[4]);
      q_scan (q1);
      fscanf (stdin, "%i", i);
    }
  else
  if (argc == 2)
    {
      *i    = atoi (argv[1]);
      q_scan (q1);
      q_scan (q2);
    }

  else
    {
      q_scan (q1);
      q_scan (q2);
      fscanf (stdin, "%i", i); 
   }
}

void
seed_random ()
{
  srand48((unsigned)time(NULL));
}

void
q_random_01 (double q1[])
{
  q1[0] = drand48 ();
  q1[1] = drand48 ();
  q1[2] = drand48 ();
  q1[3] = drand48 ();
}

void
q_random_011 (double q1[])
{
  q1[0] = drand48 ();
  q1[1] = drand48 () * 2 - 1;
  q1[2] = drand48 () * 2 - 1;
  q1[3] = drand48 () * 2 - 1;
}

void
q_random_11 (double q1[])
{
  q1[0] = drand48 () * 2 - 1;
  q1[1] = drand48 () * 2 - 1;
  q1[2] = drand48 () * 2 - 1;
  q1[3] = drand48 () * 2 - 1;
}

void
q_copy (double q1[], double q_out[])
{
  q_out[0] = q1[0];
  q_out[1] = q1[1];
  q_out[2] = q1[2];
  q_out[3] = q1[3];
}

void
q_dup (double q1[], double q_out[])
{
  q_out[0] = q1[0];
  q_out[1] = q1[1];
  q_out[2] = q1[2];
  q_out[3] = q1[3];
}


/*  Unary operators: parts */
void
q_scalar (double q1[], double q_out[])
{
  q_out[0] = q1[0];
  q_out[1] = 0.0;
  q_out[2] = 0.0;
  q_out[3] = 0.0;
}

void
q_vector (double q1[], double q_out[])
{
  q_out[0] = 0.0;
  q_out[1] = q1[1];
  q_out[2] = q1[2];
  q_out[3] = q1[3];
}

/*  Unary operators: sizes */
void
q_abs (double q1[], double q_out[])
{
  q_out[0] = sqrt(q1[0] * q1[0] + q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3]);
  q_out[1] = 0.0;
  q_out[2] = 0.0;
  q_out[3] = 0.0;
}

void
d_abs (double q1[], double *d_out)
{
  *d_out = sqrt(q1[0] * q1[0] + q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3]);
}

void
q_abs_vector (double q1[], double q_out[])
{
  q_out[0] = sqrt(q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3]);
  q_out[1] = 0.0;
  q_out[2] = 0.0;
  q_out[3] = 0.0;
}

void
d_abs_vector (double q1[], double *d_out)
{
  *d_out = sqrt(q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3]);
}

void
q_norm (double q1[], double q_out[])
{
  q_out[0] = q1[0] * q1[0] + q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3];
  q_out[1] = 0.0;
  q_out[2] = 0.0;
  q_out[3] = 0.0;
}

void
d_norm (double q1[], double *d_out)
{
  *d_out = q1[0] * q1[0] + q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3];
}

void
q_norm_vector (double q1[], double q_out[])
{
  q_out[0] = q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3];
  q_out[1] = 0.0;
  q_out[2] = 0.0;
  q_out[3] = 0.0;
}

void
d_norm_vector (double q1[], double *d_out)
{
  *d_out = q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3];
}

void
q_det (double q1[], double q_out[])
{
  double norm;
  norm = q1[0] * q1[0] + q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3];
  q_out[0] = norm * norm;
  q_out[1] = 0.0;
  q_out[2] = 0.0;
  q_out[3] = 0.0;
}

void
d_det (double q1[], double *d_out)
{
  double norm;
  norm = q1[0] * q1[0] + q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3];
  *d_out = norm * norm;
}

void
q_tau_squared (double q1[], double q_out[])
{
  q_out[0] = q1[0] * q1[0] - q1[1] * q1[1] - q1[2] * q1[2] - q1[3] * q1[3];
  q_out[1] = 0.0;
  q_out[2] = 0.0;
  q_out[3] = 0.0;
}

void
d_tau_squared (double q1[], double *d_out)
{
  *d_out = q1[0] * q1[0] - q1[1] * q1[1] - q1[2] * q1[2] - q1[3] * q1[3];
}

/*  Unary operators: automorphisms */
void
q_identity (double q1[], double q_out[])
{
  q_out[0] = q1[0];
  q_out[1] = q1[1];
  q_out[2] = q1[2];
  q_out[3] = q1[3];
}

void
q_conj (double q1[], double q_out[])
{
  q_out[0] =      q1[0];
  q_out[1] = -1 * q1[1];
  q_out[2] = -1 * q1[2];
  q_out[3] = -1 * q1[3];
}

void
q_conj_1 (double q1[], double q_out[])
{
  q_out[0] = -1 * q1[0];
  q_out[1] =      q1[1];
  q_out[2] = -1 * q1[2];
  q_out[3] = -1 * q1[3];
}

void
q_conj_2 (double q1[], double q_out[])
{
  q_out[0] = -1 * q1[0];
  q_out[1] = -1 * q1[1];
  q_out[2] =      q1[2];
  q_out[3] = -1 * q1[3];
}

void
q_inv (double q1[], double q_out[])
{
  double norm_inv;

  norm_inv =  1/(q1[0] * q1[0] + q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3]);

  q_out[0] = q1[0] * norm_inv;
  q_out[1] = -q1[1] * norm_inv;
  q_out[2] = -q1[2] * norm_inv;
  q_out[3] = -q1[3] * norm_inv;
}

void
q_adj (double q1[], double q_out[])
{
  double norm;

  norm =  q1[0] * q1[0] + q1[1] * q1[1] + q1[2] * q1[2] + q1[3] * q1[3];

  q_out[0] = q1[0] * norm;
  q_out[1] = -q1[1] * norm;
  q_out[2] = -q1[2] * norm;
  q_out[3] = -q1[3] * norm;
}


/*  Unary operators: trig */
/* sin(q) = (sin(t) cosh(|X|), cos(t) sinh(|X|) X/|X|)"> */
void
q_sin (double q1[], double q_out[])
{
  double C = 0;
  double abs_v = 0;

  d_abs_vector(q1, &abs_v);


  if (abs_v == 0) 
  {
    q_out[0] = sin(q1[0]);
    q_out[1] = 0;
    q_out[2] = 0;
    q_out[3] = 0;
    return;
  }

  C = cos(q1[0]) * sinh(abs_v) / abs_v;

  q_out[0] = sin(q1[0]) * cosh(abs_v);
  q_out[1] = C * q1[1];
  q_out[2] = C * q1[2];
  q_out[3] = C * q1[3];
}

/* cos(q) = (cos(t) cosh(|X|), sin(t) sinh(|X|) X/|X|)"> */
void
q_cos (double q1[], double q_out[])
{
  double C = 0;
  double abs_v = 0;

  d_abs_vector(q1, &abs_v);


  if (abs_v == 0) 
  {
    q_out[0] = cos(q1[0]);
    q_out[1] = 0;
    q_out[2] = 0;
    q_out[3] = 0;
    return;
  }
 
  C = - sin(q1[0]) * sinh(abs_v) / abs_v;

  q_out[0] = cos (q1[0]) * cosh (abs_v);
  q_out[1] = C * q1[1];
  q_out[2] = C * q1[2];
  q_out[3] = C * q1[3];
}

void
q_tan (double q1[], double q_out[])
{
  double abs_v = 0;
  double temp1[4], temp2[4];  

  d_abs_vector(q1, &abs_v);


  if (abs_v == 0) 
  {
    q_out[0] = tan (q1[0]);
    q_out[1] = 0;
    q_out[2] = 0;
    q_out[3] = 0;
    return;
  }

  q_sin (q1, temp1);
  q_cos (q1, temp2); 
  q_x_inv (temp1, temp2, q_out);
}

/* sinh(q) = (sinh(t) cos(|X|), cosh(t) sin(|X|) X/|X|)"> */
void
q_sinh (double q1[], double q_out[])
{
  double C = 0;
  double abs_v = 0;

  d_abs_vector(q1, &abs_v);
  
 
  if (abs_v == 0)
  {
    q_out[0] = sinh(q1[0]);
    q_out[1] = 0;
    q_out[2] = 0;
    q_out[3] = 0;
    return;
  }
 
  C = cosh(q1[0]) * sin(abs_v) / abs_v;

  q_out[0] = sinh(q1[0]) * cos(abs_v);
  q_out[1] = C * q1[1];
  q_out[2] = C * q1[2];
  q_out[3] = C * q1[3];
}

/* cosh(q) = (cosh(t) cos(|X|), sinh(t) sin(|X|) X/|X|)"> */
void
q_cosh (double q1[], double q_out[])
{
  double C = 0;
  double abs_v = 0;

  d_abs_vector(q1, &abs_v);
  

  if (abs_v == 0) 
  {
    q_out[0] = cosh(q1[0]);
    q_out[1] = 0;
    q_out[2] = 0;
    q_out[3] = 0;
    return;
  }
 
  C = sinh(q1[0]) * sin(abs_v) / abs_v;

  q_out[0] = cosh(q1[0]) * cos(abs_v);
  q_out[1] = C * q1[1];
  q_out[2] = C * q1[2];
  q_out[3] = C * q1[3];
}

void
q_tanh (double q1[], double q_out[])
{
  double abs_v = 0;
  double temp1[4], temp2[4];
 
  d_abs_vector(q1, &abs_v);

  if (abs_v == 0) 
  {
    q_out[0] = tanh (q1[0]);
    q_out[1] = 0;
    q_out[2] = 0;
    q_out[3] = 0;
    return;
  }

  q_sinh (q1, temp1);
  q_cosh (q1, temp2); 
  q_x_inv (temp1, temp2, q_out);
}

/* asin(q) = -V/|V| asinh(q V/|V|) */
void
q_asin (double q1[], double q_out[])
{
  double abs_v = 0;
  double temp1[4], temp2[4], temp3[4];
  
  d_abs_vector(q1, &abs_v);
  
  if (abs_v == 0) 
  {
    /* arbitrary, yukko */
    temp1[0] = 0;
    temp1[1] = 1;
    temp1[2] = 0;
    temp1[3] = 0;
  }
  else
  {
    q_x_scalar (q1, 1/abs_v, temp1);
  }

  q_vector (temp1, temp2);
  q_x (q1, temp2, temp3);
  q_asinh (temp3, temp1);
  q_conj (temp2, temp3);
  q_x (temp3, temp1, q_out);
}

/* acos(q) = -V/|V| acosh(q V/|V|) */
void
q_acos (double q1[], double q_out[])
{
  double abs_v = 0;
  double temp1[4], temp2[4], temp3[4];
  
  d_abs_vector(q1, &abs_v);
  
  if (abs_v == 0) 
  {
    /* arbitrary, yukko */
    temp1[0] = 0;
    temp1[1] = 1;
    temp1[2] = 0;
    temp1[3] = 0;
  }
  else
  {
    q_x_scalar (q1, -1/abs_v, temp1);
  }

  q_vector (temp1, temp2);
  q_acosh (q1, temp3);
  q_x (temp2, temp3, q_out);
}

void
q_atan (double q1[], double q_out[])
{
  double abs_v = 0;
  double temp1[4], temp2[4], temp3[4];
  
  d_abs_vector(q1, &abs_v);
  
  if (abs_v == 0) 
  {
    /* arbitrary, yukko */
    temp1[0] = 0;
    temp1[1] = 1;
    temp1[2] = 0;
    temp1[3] = 0;
  }
  else
  {
    q_x_scalar (q1, 1/abs_v, temp1);
  }

  q_vector (temp1, temp2);
  q_x (q1, temp2, temp3);
  q_atanh (temp3, temp1);
  q_conj (temp2, temp3);
  q_x (temp3, temp1, q_out);
}


/* asinh(q) = ln(q + (q^2 + 1)^.5 */
void
q_asinh (double q1[], double q_out[])
{
  double temp1[4], temp2[4], temp3[4];
  double abs_v = 0;
  
  d_abs_vector(q1, &abs_v);

  if (abs_v == 0) 
  {
    q_out[0] = asinh (q1[0]);
    q_out[1] = 0;
    q_out[2] = 0;
    q_out[3] = 0;
    return;
  }

  q_x (q1, q1, temp1);
  q_one (temp2);
  q_add (temp1, temp2, temp3);
  q_to_n (temp3, .5, temp1);
  q_add (q1, temp1, temp2);
  q_ln (temp2, q_out);
}

/* acosh(q) = ln(q +/- (q^2 - 1)^.5) */
void
q_acosh (double q1[], double q_out[])
{
  double temp1[4], temp2[4], temp3[4], sum[4], dif[4];
  double norm_sum, norm_dif;
  double abs_v = 0;
  
  d_abs_vector(q1, &abs_v);

  if (abs_v == 0) 
  {
    q_out[0] = acosh (q1[0]);
    q_out[1] = 0;
    q_out[2] = 0;
    q_out[3] = 0;
    return;
  }

  q_x (q1, q1, temp1);
  q_one (temp2);
  q_sub (temp1, temp2, temp3);
  q_to_n (temp3, .5, temp1);
  q_add (q1, temp1, sum);
  q_sub (q1, temp1, dif);

  d_norm (sum, &norm_sum);
  d_norm (dif, &norm_dif);
  if (norm_sum > norm_dif)
    {
    q_ln (sum, q_out);
    }
  else
    {
    q_ln (dif, q_out);
    }
}

/* atanh(q) = 0.5 ln((1 + q)/(1 - q)) */
void
q_atanh (double q1[], double q_out[])
{
  double temp1[4], temp2[4], temp3[4], one[4];
  double abs_v = 0;
  
  d_abs_vector(q1, &abs_v);

  q_one (one);
  q_add (one, q1, temp1);
  q_sub (one, q1, temp2);
  q_x_inv (temp1, temp2, temp3);
  q_ln (temp3, temp1);
  q_x_scalar (temp1, .5, q_out);

  /* feels like a hack */
  if (abs_v == 0 && q1[0] > 1) 
  {
    q_out[1] *= -1;
  }
}


/*  Unary operators: exponentials */

/* exp(q) = (exp(t) cos(|X|), exp(t) sin(|X|) X/|X|)"> */
void
q_exp (double q1[], double q_out[])
{
  double C = 0;
  double abs_v = 0;
  
  d_abs_vector(q1, &abs_v);

  if (abs_v == 0) 
  {
    q_out[0] = exp(q1[0]);
    q_out[1] = 0;
    q_out[2] = 0;
    q_out[3] = 0;
    return;
  }
 
  C = exp(q1[0]) * sin(abs_v) / abs_v;

  q_out[0] = exp(q1[0]) * cos(abs_v);
  q_out[1] = C * q1[1];
  q_out[2] = C * q1[2];
  q_out[3] = C * q1[3];
}

/* ln(q) = (0.5 ln t^2 + V.V, atan2(|X|, t) X/|X|)"> */
void
q_ln (double q1[], double q_out[])
{
  double C = 0;
  double abs_v = 0;

  d_abs_vector(q1, &abs_v);

  if (abs_v == 0) 
  {
    if (q1[0] > 0) {
        q_out[0] = log(q1[0]);
        q_out[1] = 0;
        q_out[2] = 0;
        q_out[3] = 0;
        return;
    }

    /* wish I knew why this was valid - actually I think it is not */
    else {
        q_out[0] = log(-1 * q1[0]);
        q_out[1] = M_PI;
        q_out[2] = 0;
        q_out[3] = 0;
	return;
    }
  }
 
  C = atan2(abs_v, q1[0]) / abs_v;

  q_out[0] = .5 * log(q1[0] * q1[0] + abs_v * abs_v);
  q_out[1] = C * q1[1];
  q_out[2] = C * q1[2];
  q_out[3] = C * q1[3];
}

void 
q_to_n (double q1[], double d, double q_out[])
{
  double temp1[4], temp2[4];

  q_ln (q1, temp1);
  q_x_scalar (temp1, d, temp2);
  q_exp (temp2, q_out);
}


/*  Binary operators: +/- */
void
q_add_scalar (double q1[], double d, double q_out[])
{
  q_out[0] = q1[0] + d;
  q_out[1] = q1[1];
  q_out[2] = q1[2];
  q_out[3] = q1[3];
}

void
q_add (double q1[], double q2[], double q_out[])
{
  q_out[0] = q1[0] + q2[0];
  q_out[1] = q1[1] + q2[1];
  q_out[2] = q1[2] + q2[2];
  q_out[3] = q1[3] + q2[3];
}

void
q_sub (double q1[], double q2[], double q_out[])
{
  q_out[0] = q1[0] - q2[0];
  q_out[1] = q1[1] - q2[1];
  q_out[2] = q1[2] - q2[2];
  q_out[3] = q1[3] - q2[3];
}


/*  Binary operators: multiplication/division */
void
q_x_scalar (double q1[], double d, double q_out[])
{
  q_out[0] = q1[0] * d;
  q_out[1] = q1[1] * d;
  q_out[2] = q1[2] * d;
  q_out[3] = q1[3] * d;
}

void
q_x (double q1[], double q2[], double q_out[])
{
  q_out[0] = q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2] - q1[3] * q2[3];
  q_out[1] = q1[0] * q2[1] + q1[1] * q2[0] + q1[2] * q2[3] - q1[3] * q2[2];
  q_out[2] = q1[0] * q2[2] + q1[2] * q2[0] + q1[3] * q2[1] - q1[1] * q2[3];
  q_out[3] = q1[0] * q2[3] + q1[3] * q2[0] + q1[1] * q2[2] - q1[2] * q2[1];
}

void
q_x_even (double q1[], double q2[], double q_out[])
{
  q_out[0] = q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2] - q1[3] * q2[3];
  q_out[1] = q1[0] * q2[1] + q1[1] * q2[0];
  q_out[2] = q1[0] * q2[2] + q1[2] * q2[0];
  q_out[3] = q1[0] * q2[3] + q1[3] * q2[0];
}

void
q_x_odd (double q1[], double q2[], double q_out[])
{
  q_out[0] = 0;
  q_out[1] = q1[2] * q2[3] - q1[3] * q2[2];
  q_out[2] = q1[3] * q2[1] - q1[1] * q2[3];
  q_out[3] = q1[1] * q2[2] - q1[2] * q2[1];
}

void
q_x_inv (double q1[], double q2[], double q_out[])
{
  double norm_inv = 0;
 
  d_norm(q2, &norm_inv);

  norm_inv = 1/norm_inv;

  q_out[0] = (q1[0] * q2[0] + q1[1] * q2[1] + q1[2] * q2[2] + q1[3] * q2[3]) * norm_inv;
  q_out[1] = (-q1[0] * q2[1] + q1[1] * q2[0] - q1[2] * q2[3] + q1[3] * q2[2]) * norm_inv;
  q_out[2] = (-q1[0] * q2[2] + q1[2] * q2[0] - q1[3] * q2[1] + q1[1] * q2[3]) * norm_inv;
  q_out[3] = (-q1[0] * q2[3] + q1[3] * q2[0] - q1[1] * q2[2] + q1[2] * q2[1]) * norm_inv;
}

void
q_scan (double q1[])
{
  fscanf (stdin, "%lf %lf %lf %lf", &q1[0], &q1[1], &q1[2], &q1[3]);
}

void
q_print (double q1[])
{
  fprintf (stdout, "%f %f %f %f\n", q1[0], q1[1], q1[2], q1[3]);
}


/*  This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
